@@ -15,14 +15,20 @@ Brightfield
     Images coherently illuminated samples.
 """
 
-import numpy as np
-from .features import Feature, StructuralFeature
-from .image import Image, pad_image_to_fft
-from .types import ArrayLike, PropertyLike
 
-import scipy.ndimage
+from pint.quantity import Quantity
+from deeptrack.backend.units import ConversionTable
+from deeptrack.properties import propagate_data_to_dependencies
+import numpy as np
+from .features import DummyFeature, Feature, StructuralFeature
+from .image import Image, pad_image_to_fft, maybe_cupy
+from .types import ArrayLike, PropertyLike
+from .backend._config import cupy
 from scipy.ndimage import convolve
 
+from . import units as u
+from deeptrack import image
+
 
 class Microscope(StructuralFeature):
     """Image a sample using an optical system.
@@ -40,16 +46,19 @@ class Microscope(StructuralFeature):
     __distributed__ = False
 
     def __init__(self, sample: Feature, objective: Feature, **kwargs):
-        super().__init__(sample=sample, objective=objective, **kwargs)
+        super().__init__(**kwargs)
+        self._sample = self.add_feature(sample)
+        self._objective = self.add_feature(objective)
 
-    def get(self, image, sample, objective, **kwargs):
+    def get(self, image, **kwargs):
 
         # Grab properties from the objective to pass to the sample
-        new_kwargs = objective.properties.current_value_dict(**kwargs)
-        new_kwargs.update(kwargs)
-        kwargs = new_kwargs
+        additional_sample_kwargs = self._objective.properties()
+        propagate_data_to_dependencies(self._sample, **additional_sample_kwargs)
+
+        with u.context("dt", pixel_size=additional_sample_kwargs["voxel_size"][0]):
+            list_of_scatterers = self._sample()
 
-        list_of_scatterers = sample.resolve(**kwargs)
         if not isinstance(list_of_scatterers, list):
             list_of_scatterers = [list_of_scatterers]
 
@@ -64,33 +73,21 @@ class Microscope(StructuralFeature):
             if scatterer.get_property("is_field", default=False)
         ]
 
-        sample_volume, limits = _create_volume(volume_samples, **kwargs)
+        sample_volume, limits = _create_volume(
+            volume_samples, **additional_sample_kwargs
+        )
         sample_volume = Image(sample_volume)
 
         for scatterer in volume_samples + field_samples:
             sample_volume.merge_properties_from(scatterer)
 
-        imaged_sample = objective.resolve(
-            sample_volume, limits=limits, fields=field_samples, **kwargs
+        propagate_data_to_dependencies(
+            self._objective,
+            limits=limits,
+            fields=field_samples,
         )
 
-        upscale = kwargs["upscale"]
-        shape = imaged_sample.shape
-        if upscale > 1:
-            mean_imaged_sample = np.reshape(
-                imaged_sample,
-                (
-                    shape[0] // upscale,
-                    upscale,
-                    shape[1] // upscale,
-                    upscale,
-                    shape[2],
-                ),
-            ).mean(axis=(3, 1))
-
-            imaged_sample = Image(mean_imaged_sample).merge_properties_from(
-                imaged_sample
-            )
+        imaged_sample = self._objective.resolve(sample_volume)
 
         # Merge with input
         if not image:
@@ -102,15 +99,6 @@ class Microscope(StructuralFeature):
             image[i].merge_properties_from(imaged_sample)
         return image
 
-    def _update(self, **kwargs):
-        self.properties["sample"].update(
-            **{
-                **kwargs,
-                **self.objective.update(**kwargs).current_value.properties,
-            }
-        )
-        super()._update(**kwargs)
-
 
 # OPTICAL SYSTEMS
 
@@ -133,8 +121,6 @@ class Optics(Feature):
         included to define the resolution in the z-direction.
     refractive_index_medium : float
         The refractive index of the medium.
-    upscale : int
-        Upscales the pupil function for a more accurate result.
     padding : array_like[int, int, int, int]
         Pads the sample volume with zeros to avoid edge effects.
     output_region : array_like[int, int, int, int]
@@ -146,28 +132,36 @@ class Optics(Feature):
 
     """
 
+    __conversion_table__ = ConversionTable(
+        wavelength=(u.meter, u.meter),
+        resolution=(u.meter, u.meter),
+        voxel_size=(u.meter, u.meter),
+    )
+
     def __init__(
         self,
         NA: PropertyLike[float] = 0.7,
         wavelength: PropertyLike[float] = 0.66e-6,
         magnification: PropertyLike[float] = 10,
-        resolution: PropertyLike[float or ArrayLike[float]] = (1e-6, 1e-6, 1e-6),
+        resolution: PropertyLike[float or ArrayLike[float]] = 1e-6,
         refractive_index_medium: PropertyLike[float] = 1.33,
-        upscale: PropertyLike[float] = 1,
         padding: PropertyLike[ArrayLike[int]] = (10, 10, 10, 10),
         output_region: PropertyLike[ArrayLike[int]] = (0, 0, 128, 128),
         pupil: Feature = None,
+        illumination: Feature = None,
         **kwargs
     ):
-        def get_voxel_size(resolution, magnification, upscale):
-            if (
-                not isinstance(resolution, (list, tuple, np.ndarray))
-                or len(resolution) == 1
-            ):
-                resolution = (resolution,) * 3
-            elif len(resolution) == 2:
-                resolution = (*resolution, np.min(resolution))
-            return np.array(resolution) / (magnification * upscale)
+        def get_voxel_size(resolution, magnification):
+            props = self._normalize(resolution=resolution, magnification=magnification)
+            return np.ones((3,)) * props["resolution"] / props["magnification"]
+
+        def get_pixel_size(resolution, magnification):
+            props = self._normalize(resolution=resolution, magnification=magnification)
+            pixel_size = props["resolution"] / props["magnification"]
+            if isinstance(pixel_size, Quantity):
+                return pixel_size.to(u.meter).magnitude
+            else:
+                return pixel_size
 
         super().__init__(
             NA=NA,
@@ -175,24 +169,19 @@ class Optics(Feature):
             refractive_index_medium=refractive_index_medium,
             magnification=magnification,
             resolution=resolution,
-            upscale=upscale,
             padding=padding,
             output_region=output_region,
-            upscaled_output_region=lambda output_region, upscale: [
-                i * upscale for i in output_region
-            ],
-            pupil=pupil,
-            voxel_size=lambda resolution, magnification, upscale: get_voxel_size(
-                resolution, magnification, upscale
-            ),
+            voxel_size=get_voxel_size,
+            pixel_size=get_pixel_size,
+            limits=None,
+            fields=None,
             **kwargs
         )
 
-    def _process_and_get(self, image, **kwargs):
-        output = super()._process_and_get(image, **kwargs)
-        pupil = self._pupil(output[-1].shape[:2], defocus=[0], **kwargs)[0]
-
-        return output, {"pupil_at_focus": pupil}
+        self.pupil = self.add_feature(pupil) if pupil else DummyFeature()
+        self.illumination = (
+            self.add_feature(illumination) if illumination else DummyFeature()
+        )
 
     def _pupil(
         self,
@@ -201,9 +190,6 @@ class Optics(Feature):
         wavelength,
         refractive_index_medium,
         voxel_size,
-        upscale,
-        pupil,
-        aberration=None,
         include_aberration=True,
         defocus=0,
         **kwargs
@@ -221,80 +207,71 @@ class Optics(Feature):
         y = (np.linspace(-(shape[1] / 2), shape[1] / 2 - 1, shape[1])) / y_radius + 1e-8
 
         W, H = np.meshgrid(y, x)
+        W = maybe_cupy(W)
+        H = maybe_cupy(H)
         RHO = W ** 2 + H ** 2
-        RHO[RHO > 1] = 1
-        pupil_function = ((RHO < 1) * 1.0).astype(np.complex)
+        pupil_function = Image((RHO < 1) + 0.0j, copy=False)
         # Defocus
-        z_shift = (
+        z_shift = Image(
             2
             * np.pi
             * refractive_index_medium
             / wavelength
             * voxel_size[2]
-            * np.sqrt(1 - (NA / refractive_index_medium) ** 2 * RHO)
+            * np.sqrt(1 - (NA / refractive_index_medium) ** 2 * RHO),
+            copy=False,
         )
 
-        # Downsample the upsampled pupil
+        try:
+            z_shift = np.nan_to_num(z_shift, False, 0, 0, 0)
+        except TypeError:
+            np.nan_to_num(z_shift, z_shift)
 
-        pupil_function[np.isnan(pupil_function)] = 0
-        pupil_function[np.isinf(pupil_function)] = 0
-        pupil_function_is_nonzero = pupil_function != 0
+        defocus = np.reshape(defocus, (-1, 1, 1))
+        z_shift = defocus * np.expand_dims(z_shift, axis=0)
 
         if include_aberration:
-            pupil = pupil or aberration
+            pupil = self.pupil
             if isinstance(pupil, Feature):
-                pupil_function = pupil.resolve(pupil_function, **kwargs)
+
+                pupil_function = pupil(pupil_function)
             elif isinstance(pupil, np.ndarray):
                 pupil_function *= pupil
 
-        pupil_functions = []
-        for z in defocus:
-            pupil_at_z = Image(pupil_function)
-            pupil_at_z[pupil_function_is_nonzero] *= np.exp(
-                1j * z_shift[pupil_function_is_nonzero] * z
-            )
-            pupil_functions.append(pupil_at_z)
+        pupil_functions = pupil_function * np.exp(1j * z_shift)
 
         return pupil_functions
 
     def _pad_volume(
-        self, volume, limits=None, padding=None, upscaled_output_region=None, **kwargs
+        self, volume, limits=None, padding=None, output_region=None, **kwargs
     ):
         if limits is None:
             limits = np.zeros((3, 2))
 
         new_limits = np.array(limits)
-        upscaled_output_region = np.array(upscaled_output_region)
+        output_region = np.array(output_region)
 
         # Replace None entries with current limit
-        upscaled_output_region[0] = (
-            upscaled_output_region[0]
-            if not upscaled_output_region[0] is None
-            else new_limits[0, 0]
+        output_region[0] = (
+            output_region[0] if not output_region[0] is None else new_limits[0, 0]
         )
-        upscaled_output_region[1] = (
-            upscaled_output_region[1]
-            if not upscaled_output_region[1] is None
-            else new_limits[0, 1]
+        output_region[1] = (
+            output_region[1] if not output_region[1] is None else new_limits[0, 1]
         )
-        upscaled_output_region[2] = (
-            upscaled_output_region[2]
-            if not upscaled_output_region[2] is None
-            else new_limits[1, 0]
+        output_region[2] = (
+            output_region[2] if not output_region[2] is None else new_limits[1, 0]
         )
-        upscaled_output_region[3] = (
-            upscaled_output_region[3]
-            if not upscaled_output_region[3] is None
-            else new_limits[1, 1]
+        output_region[3] = (
+            output_region[3] if not output_region[3] is None else new_limits[1, 1]
         )
 
         for i in range(2):
             new_limits[i, :] = (
-                np.min([new_limits[i, 0], upscaled_output_region[i] - padding[1]]),
+                np.min([new_limits[i, 0], output_region[i] - padding[1]]),
                 np.max(
                     [
                         new_limits[i, 1],
-                        upscaled_output_region[i + 2] + padding[i + 2],
+                        output_region[i + 2] + padding[i + 2],
                     ]
                 ),
             )
@@ -318,6 +295,7 @@ class Optics(Feature):
 
 
 class Fluorescence(Optics):
+
     """Optical device for fluorescenct imaging
 
     Images samples by creating a discretized volume, where each pixel
@@ -337,8 +315,6 @@ class Fluorescence(Optics):
         included to define the resolution in the z-direction.
     refractive_index_medium : float
         The refractive index of the medium.
-    upscale : int
-        Upscales the pupil function for a more accurate result.
     padding : array_like[int, int, int, int]
         Pads the sample volume with zeros to avoid edge effects.
     output_region : array_like[int, int, int, int]
@@ -350,8 +326,11 @@ class Fluorescence(Optics):
 
     """
 
+    __gpu_compatible__ = True
+
     def get(self, illuminated_volume, limits, **kwargs):
         """Convolves the image with a pupil function"""
+
         # Pad volume
         padded_volume, limits = self._pad_volume(
             illuminated_volume, limits=limits, **kwargs
@@ -359,9 +338,7 @@ class Fluorescence(Optics):
 
         # Extract indexes of the output region
         pad = kwargs.get("padding", (0, 0, 0, 0))
-        output_region = np.array(
-            kwargs.get("upscaled_output_region", (None, None, None, None))
-        )
+        output_region = np.array(kwargs.get("output_region", (None, None, None, None)))
         output_region[0] = (
             None
             if output_region[0] is None
@@ -390,7 +367,9 @@ class Fluorescence(Optics):
         ]
         z_limits = limits[2, :]
 
-        output_image = Image(np.zeros((*padded_volume.shape[0:2], 1)))
+        output_image = Image(
+            maybe_cupy(np.zeros((*padded_volume.shape[0:2], 1))), copy=False
+        )
 
         index_iterator = range(padded_volume.shape[2])
 
@@ -405,40 +384,41 @@ class Fluorescence(Optics):
         z_values = z_iterator[~zero_plane]
 
         # Further pad image to speed up fft
-        volume = pad_image_to_fft(padded_volume, axes=(0, 1))
+        volume = maybe_cupy(pad_image_to_fft(padded_volume, axes=(0, 1)))
 
         pupils = self._pupil(volume.shape[:2], defocus=z_values, **kwargs)
-        pupil_iterator = iter(pupils)
+
+        z_index = 0
 
         # Loop through voluma and convole sample with pupil function
+
         for i, z in zip(index_iterator, z_iterator):
 
             if zero_plane[i]:
                 continue
 
-            image = volume[:, :, i]
-            pupil = Image(next(pupil_iterator))
+            pupil = pupils[z_index]
+            z_index += 1
 
             psf = np.square(np.abs(np.fft.ifft2(np.fft.fftshift(pupil))))
+
             optical_transfer_function = np.fft.fft2(psf)
 
-            fourier_field = np.fft.fft2(image)
+            fourier_field = np.fft.fft2(volume[:, :, i])
             convolved_fourier_field = fourier_field * optical_transfer_function
 
-            field = Image(np.fft.ifft2(convolved_fourier_field))
+            field = np.fft.ifft2(convolved_fourier_field)
 
-            # Discard remaining imaginary part (should be 0 up to rounding error)
+            # # Discard remaining imaginary part (should be 0 up to rounding error)
             field = np.real(field)
 
-            output_image[:, :, 0] += field[
+            output_image._value[:, :, 0] += field[
                 : padded_volume.shape[0], : padded_volume.shape[1]
             ]
 
         output_image = output_image[pad[0] : -pad[2], pad[1] : -pad[3]]
-        try:
-            output_image.properties = illuminated_volume.properties + pupil.properties
-        except UnboundLocalError:
-            output_image.properties = illuminated_volume.properties
+
+        output_image.properties = illuminated_volume.properties + pupils.properties
 
         return output_image
 
@@ -468,8 +448,6 @@ class Brightfield(Optics):
         included to define the resolution in the z-direction.
     refractive_index_medium : float
         The refractive index of the medium.
-    upscale : int
-        Upscales the pupil function for a more accurate result.
     padding : array_like[int, int, int, int]
         Pads the sample volume with zeros to avoid edge effects.
     output_region : array_like[int, int, int, int]
@@ -481,6 +459,8 @@ class Brightfield(Optics):
 
     """
 
+    __gpu_compatible__ = True
+
     def get(self, illuminated_volume, limits, fields, **kwargs):
         """Convolves the image with a pupil function"""
         # Pad volume
@@ -490,9 +470,7 @@ class Brightfield(Optics):
 
         # Extract indexes of the output region
         pad = kwargs.get("padding", (0, 0, 0, 0))
-        output_region = np.array(
-            kwargs.get("upscaled_output_region", (None, None, None, None))
-        )
+        output_region = np.array(kwargs.get("output_region", (None, None, None, None)))
         output_region[0] = (
             None
             if output_region[0] is None
@@ -521,7 +499,7 @@ class Brightfield(Optics):
         ]
         z_limits = limits[2, :]
 
-        output_image = Image(np.zeros((*padded_volume.shape[0:2], 1)))
+        output_image = Image(image.maybe_cupy(np.zeros((*padded_volume.shape[0:2], 1))))
 
         index_iterator = range(padded_volume.shape[2])
         z_iterator = np.linspace(
@@ -538,21 +516,23 @@ class Brightfield(Optics):
 
         voxel_size = kwargs["voxel_size"]
 
-        pupils = self._pupil(
-            volume.shape[:2], defocus=[1], include_aberration=False, **kwargs
-        ) + self._pupil(
-            volume.shape[:2], defocus=[-z_limits[1]], include_aberration=True, **kwargs
-        )
+        pupils = [
+            self._pupil(
+                volume.shape[:2], defocus=[1], include_aberration=False, **kwargs
+            )[0],
+            self._pupil(
+                volume.shape[:2],
+                defocus=[-z_limits[1]],
+                include_aberration=True,
+                **kwargs
+            )[0],
+        ]
 
         pupil_step = np.fft.fftshift(pupils[0])
 
-        if "illumination" in kwargs:
-            light_in = np.ones(volume.shape[:2], dtype=np.complex)
-            light_in = kwargs["illumination"].resolve(light_in, **kwargs)
-            light_in = np.fft.fft2(light_in)
-        else:
-            light_in = np.zeros(volume.shape[:2], dtype=np.complex)
-            light_in[0, 0] = light_in.size
+        light_in = image.maybe_cupy(np.ones(volume.shape[:2], dtype=np.complex))
+        light_in = self.illumination.resolve(light_in)
+        light_in = np.fft.fft2(light_in)
 
         K = 2 * np.pi / kwargs["wavelength"]
 
@@ -572,6 +552,7 @@ class Brightfield(Optics):
                         include_aberration=False,
                         **kwargs
                     )[0]
+
                     propagation_matrix = propagation_matrix * np.exp(
                         1j
                         * voxel_size[-1]
@@ -608,6 +589,7 @@ class Brightfield(Optics):
                 include_aberration=False,
                 **kwargs
             )[0]
+
             propagation_matrix = propagation_matrix * np.exp(
                 -1j
                 * voxel_size[-1]
@@ -689,8 +671,9 @@ def _get_position(image, mode="corner", return_z=False):
     num_outputs = 2 + return_z
 
     if mode == "corner" and image.size > 0:
+        import scipy.ndimage
 
-        shift = scipy.ndimage.measurements.center_of_mass(np.abs(np.array(image)))
+        shift = scipy.ndimage.measurements.center_of_mass(np.abs(image))
 
         if np.isnan(shift).any():
             shift = np.array(image.shape) / 2
@@ -698,10 +681,9 @@ def _get_position(image, mode="corner", return_z=False):
     else:
         shift = np.zeros((num_outputs))
 
-    upscale = image.get_property("upscale")
-    position = np.array(image.get_property("position")) * upscale
+    position = np.array(image.get_property("position", default=None))
 
-    position[:2] = position[:2] + 0.5 * (upscale - 1)
+    # position[:2] = position[:2]
 
     if position is None:
         return position
@@ -715,14 +697,7 @@ def _get_position(image, mode="corner", return_z=False):
     elif len(position) == 2:
         if return_z:
             outp = (
-                np.array(
-                    [
-                        position[0],
-                        position[1],
-                        image.get_property("z", default=0)
-                        * image.get_property("upscale"),
-                    ]
-                )
+                np.array([position[0], position[1], image.get_property("z", default=0)])
                 - shift
             )
             return outp
@@ -735,9 +710,8 @@ def _get_position(image, mode="corner", return_z=False):
 def _create_volume(
     list_of_scatterers,
     pad=(0, 0, 0, 0),
-    upscaled_output_region=(None, None, None, None),
+    output_region=(None, None, None, None),
     refractive_index_medium=1.33,
-    upscale=1,
     **kwargs
 ):
     # Converts a list of scatterers into a volume.
@@ -748,26 +722,10 @@ def _create_volume(
     volume = np.zeros((1, 1, 1), dtype=np.complex)
     limits = None
     OR = np.zeros((4,))
-    OR[0] = (
-        np.inf
-        if upscaled_output_region[0] is None
-        else int(upscaled_output_region[0] - pad[0])
-    )
-    OR[1] = (
-        -np.inf
-        if upscaled_output_region[1] is None
-        else int(upscaled_output_region[1] - pad[1])
-    )
-    OR[2] = (
-        np.inf
-        if upscaled_output_region[2] is None
-        else int(upscaled_output_region[2] + pad[2])
-    )
-    OR[3] = (
-        -np.inf
-        if upscaled_output_region[3] is None
-        else int(upscaled_output_region[3] + pad[3])
-    )
+    OR[0] = np.inf if output_region[0] is None else int(output_region[0] - pad[0])
+    OR[1] = -np.inf if output_region[1] is None else int(output_region[1] - pad[1])
+    OR[2] = np.inf if output_region[2] is None else int(output_region[2] + pad[2])
+    OR[3] = -np.inf if output_region[3] is None else int(output_region[3] + pad[3])
 
     for scatterer in list_of_scatterers:
 
@@ -805,7 +763,8 @@ def _create_volume(
                 constant_values=0,
             )
         )
-        padded_scatterer.properties = scatterer.properties
+        padded_scatterer.merge_properties_from(scatterer)
+
         scatterer = padded_scatterer
         position = _get_position(scatterer, mode="corner", return_z=True)
         shape = np.array(scatterer.shape)
